//! Utilities for handling the query result data
//!
//! Some of the functionality provided by this module may be rendered redundant if the dependency on `prometheus-http-query` is dropped
//! in the future in favor of a fully custom implementation of the query client and the response deserialization process.

use prometheus_http_query::response::{InstantVector, RangeVector, Sample};
use rayon::prelude::*;
use serde::de::DeserializeOwned;
use std::collections::HashMap;

/// Helper trait for bulk type conversion of query result data
///
/// Convert the query result data such as `Vec<InstantVector>` and `Vec<RangeVector>`
/// containing raw labels of the type `HashMap<String, String>` into special custom types that support
/// user defined structs that mimic the structure of the labels.
///
/// The user defined struct should implement serde's `Deserialize` trait and must contain fields of
/// the same name as the keys in the labels/metric's `HashMap`. This is similar to deserializing from a __Json__ object.
/// For example, if the labels `HashMap` has keys named "foo", "bar", the struct must have the fields of the same name.
/// It would look something like `struct Custom { foo: String, bar: String }`
///
/// The [``Convert::convert``] and [``Convert::convert_par``] are both fallible, and return a `serde_json::Error` type on failure.
/// The former function processes data in a single thread, while the latter uses `rayon` for multithreaded processing.
///
/// For `InstantVector` or a `RangeVector`, use `try_into` or `try_from` using [`CustomInstantVec`] and [`CustomRangeVec`], or use the
/// free-standing function [`convert_labels`] to convert a single set of labels (HashMap).
///
/// # Example
/// ```ignore
/// use mquery::{result::Error, utils::Convert, QueryManager};
///
/// #[derive(serde::Deserialize, Debug, Clone)]
/// struct Custom { // custom type that represents the metric labels
///     method: String,
///     endpoint: String,
/// }
///
/// #[tokio::main]
/// async fn main() -> Result<(), Error> {
///     let resp = QueryManager::default().query("http_requests").await?;
///     let data = resp.into_vector().unwrap();
///     // returns `Ok(CustomInstantVec<Custom>)` if conversion succeeds
///     if let Ok(converted) = data.convert_par::<Custom>() {
///         println!("{converted:?}");
///     }
///     Ok(())
/// }
/// ```
pub trait Convert {
    type Target<M: DeserializeOwned>;
    type Error;
    /// Bulk convert the query result data into a 'Custom' type that can hold a user-defined struct of the same structure as the label data
    ///
    /// For a single set of labels, use `try_into` or `try_from` using [`CustomInstantVec`] and [`CustomRangeVec`].
    /// See the trait level docs for more information.
    fn convert<M: DeserializeOwned>(self) -> Result<Self::Target<M>, Self::Error>;

    /// Parallelly (using `rayon`) convert the query result data into a 'Custom' type that can hold a user-defined struct of the same structure as the label data
    ///
    /// For a single set of labels, use `try_into` or `try_from` using [`CustomInstantVec`] and [`CustomRangeVec`].
    /// See the trait level docs for more information.
    fn convert_par<M: DeserializeOwned + Send>(self) -> Result<Self::Target<M>, Self::Error>;
}

impl Convert for Vec<InstantVector> {
    type Target<M: DeserializeOwned> = Vec<CustomInstantVec<M>>;
    type Error = serde_json::Error;

    fn convert<M: DeserializeOwned>(self) -> Result<Self::Target<M>, Self::Error> {
        self.into_iter().map(|v| v.try_into()).collect()
    }

    fn convert_par<M: DeserializeOwned + Send>(self) -> Result<Self::Target<M>, Self::Error> {
        self.into_par_iter().map(|v| v.try_into()).collect()
    }
}

impl Convert for Vec<RangeVector> {
    type Target<M: DeserializeOwned> = Vec<CustomRangeVec<M>>;
    type Error = serde_json::Error;

    fn convert<M: DeserializeOwned>(self) -> Result<Self::Target<M>, Self::Error> {
        self.into_iter().map(|v| v.try_into()).collect()
    }

    fn convert_par<M: DeserializeOwned + Send>(self) -> Result<Self::Target<M>, Self::Error> {
        self.into_par_iter().map(|v| v.try_into()).collect()
    }
}

pub fn convert_labels<T: DeserializeOwned>(
    labels: HashMap<String, String>,
) -> Result<T, serde_json::Error> {
    serde_json::from_value(serde_json::to_value(labels)?)
}

/// Analogue of the [`InstantVector`] type which is generic over the type that represents labels
///
/// This can be used directly to convert an [`InstantVector`] using `try_from` with type hinting,
/// or use the [`Convert`] trait for bulk conversion of type `Vec<InstantVector>`.
/// See the trait level docs for [`Convert`] for more information.
#[derive(Clone, Debug)]
pub struct CustomInstantVec<M> {
    pub labels: M,
    pub sample: Sample,
}

impl<M: DeserializeOwned> TryFrom<InstantVector> for CustomInstantVec<M> {
    type Error = serde_json::Error;

    fn try_from(value: InstantVector) -> Result<Self, Self::Error> {
        let (labels, sample) = value.into_inner();
        Ok(Self {
            labels: convert_labels(labels)?,
            sample,
        })
    }
}

/// Analogue of the [`RangeVector`] type which is generic over the type that represents labels
///
/// This can be used directly to convert a [`RangeVector`] using `try_from` with type hinting,
/// or use the [`Convert`] trait for bulk conversion of type `Vec<InstantVector>`.
/// See the trait level docs for [`Convert`] for more information.
pub struct CustomRangeVec<M> {
    pub labels: M,
    pub samples: Vec<Sample>,
}

impl<M: DeserializeOwned> TryFrom<RangeVector> for CustomRangeVec<M> {
    type Error = serde_json::Error;

    fn try_from(value: RangeVector) -> Result<Self, Self::Error> {
        let (labels, samples) = value.into_inner();
        Ok(Self {
            labels: convert_labels(labels)?,
            samples,
        })
    }
}
